import {
	GetServerSideProps,
	NextApiHandler,
	PageConfig,
	NextApiRequest as DefaultNextApiRequest, PreviewData, GetServerSidePropsContext,
} from 'next';
import { ParsedUrlQuery } from 'querystring';
import { deserializeBody, EncTypeDeserializerMap } from '../utils/serialization';
import { IceformNextServerResponse } from './response';
import {
	BODY_COOKIE_KEY, CONTENT_TYPE_COOKIE_KEY,
	CookieManager,
	STATUS_CODE_COOKIE_KEY,
	STATUS_MESSAGE_COOKIE_KEY,
} from '../utils/cookies';
import {
	METHOD_FORM_KEY, METHODS_WITH_BODY,
	PREVENT_REDIRECT_FORM_KEY,
} from '../common/constants';
import {
	ACTION_STATUS_CODE,
	DEFAULT_METHOD,
} from './constants';

export const getApiConfig = (customConfig = {} as PageConfig) => ({
	api: {
		...(customConfig.api ?? {}),
		bodyParser: false,
	},
});

export type OnActionFunction<
	Props extends Record<string, unknown> = Record<string, unknown>,
	Params extends ParsedUrlQuery = ParsedUrlQuery,
	Preview extends PreviewData = PreviewData,
> = (
	actionReq: DefaultNextApiRequest,
	actionRes: IceformNextServerResponse,
	context: GetServerSidePropsContext<Params, Preview>,
) => Promise<Awaited<ReturnType<GetServerSideProps<Props, Params, Preview>>> | undefined>;

export interface ActionWrapperOptions {
	fn: NextApiHandler,
	onAction?: OnActionFunction,
	deserializers?: EncTypeDeserializerMap,
	/**
	 * Maps the Location header from the handler response to an accessible URL.
	 * @param url
	 */
	mapLocationToRedirectDestination?: (referer: string, url: string) => string,
}

export const wrapApiHandler = (
	options: ActionWrapperOptions,
): NextApiHandler => async (req, res) => {
	const reqMut = req as unknown as Record<string, unknown>;
	if (METHODS_WITH_BODY.includes(req.method?.toUpperCase() as typeof METHODS_WITH_BODY[number])) {
		reqMut.body = await deserializeBody({
			req,
			deserializers: options.deserializers,
		});
	}
	return options.fn(reqMut as unknown as DefaultNextApiRequest, res);
};

export const getServerSideProps = (
	options: ActionWrapperOptions,
): GetServerSideProps => async (ctx) => {
	const { referer = '/' } = ctx.req.headers;
	const deserialized = await deserializeBody({
		req: ctx.req,
		deserializers: options.deserializers,
	});

	const defaultMethod = ctx.req.method ?? DEFAULT_METHOD;
	let effectiveMethod = defaultMethod;
	let mockReqBody: unknown = deserialized;
	if (typeof deserialized === 'object' && deserialized !== null) {
		const {
			[METHOD_FORM_KEY]: theMethod,
			...theMockReqBody
		} = deserialized as {
			[METHOD_FORM_KEY]?: string,
			[key: string]: unknown,
		};
		effectiveMethod = theMethod ?? defaultMethod;
		mockReqBody = theMockReqBody;
	}

	const mockReq = {
		...ctx.req,
		body: mockReqBody,
		query: {
			...ctx.query,
			...(ctx.params ?? {}),
		},
		// ?: how to prevent malicious method spoofing?
		method: effectiveMethod,
	} as DefaultNextApiRequest;

	const mockRes = new IceformNextServerResponse(ctx.req);
	await options.fn(mockReq, mockRes);

	const cookieManager = new CookieManager(ctx);
	cookieManager.setCookie(STATUS_CODE_COOKIE_KEY, mockRes.statusCode.toString());
	cookieManager.setCookie(STATUS_MESSAGE_COOKIE_KEY, mockRes.statusMessage);
	if (mockRes.data) {
		cookieManager.setCookie(BODY_COOKIE_KEY, mockRes.data as string);
		if (mockRes.contentType) {
			cookieManager.setCookie(CONTENT_TYPE_COOKIE_KEY, mockRes.contentType);
		}
	}

	if (typeof options.onAction === 'function') {
		const onActionResult = await options.onAction(
			mockReq,
			mockRes,
			ctx,
		);
		if (onActionResult) {
			return onActionResult;
		}
	}

	const preventRedirect = (
		typeof mockReq.body === 'object'
		&& mockReq.body !== null
		&& PREVENT_REDIRECT_FORM_KEY in mockReq.body
	);
	const redirectDestination = (
		mockRes.location
		&& typeof options.mapLocationToRedirectDestination === 'function'
		&& !preventRedirect
	)
		? options.mapLocationToRedirectDestination(referer, mockRes.location)
		: referer;

	return {
		redirect: {
			destination: redirectDestination,
			statusCode: ACTION_STATUS_CODE,
		},
		props: {
			query: ctx.query,
			body: mockRes.data,
		},
	};
};
