import { ParsedUrlQuery } from 'querystring';
import { GetServerSideProps, GetServerSidePropsContext, PreviewData } from 'next';
import { deserialize } from 'seroval';
import { NextApiRequest, NextApiResponse } from '../common/types';
import {
	DEFAULT_ENCODING,
	DEFAULT_METHOD,
	DEFAULT_RESPONSE_STATUS_CODE,
} from './constants';
import { getBody } from '../utils/request';
import {
	BODY_COOKIE_KEY, CONTENT_TYPE_COOKIE_KEY,
	CookieManager,
	STATUS_CODE_COOKIE_KEY,
	STATUS_MESSAGE_COOKIE_KEY,
} from '../utils/cookies';
import { ENCTYPE_APPLICATION_JSON, ENCTYPE_APPLICATION_OCTET_STREAM } from '../common/enctypes';
import { METHODS_WITH_BODY } from '../common/constants';

export type DestinationGetServerSideProps<
	Props extends Record<string, unknown> = Record<string, unknown>,
	Params extends ParsedUrlQuery = ParsedUrlQuery,
	Preview extends PreviewData = PreviewData,
> = (
	actionReq: NextApiRequest,
	actionRes: NextApiResponse,
	context: GetServerSidePropsContext<Params, Preview>,
) => ReturnType<GetServerSideProps<Props, Params, Preview>>;

export interface DestinationWrapperOptions {
	fn?: DestinationGetServerSideProps;
}

export const getServerSideProps = (
	options = {} as DestinationWrapperOptions,
): GetServerSideProps => async (ctx) => {
	const req: NextApiRequest = {
		query: {
			...ctx.query,
			...(ctx.params ?? {}),
		},
		body: null,
	};
	const { method = DEFAULT_METHOD } = ctx.req;

	if (METHODS_WITH_BODY.includes(method.toUpperCase() as typeof METHODS_WITH_BODY[number])) {
		const body = await getBody(ctx.req);
		req.body = body.toString(DEFAULT_ENCODING);
	}

	const cookieManager = new CookieManager(ctx);
	// TODO how to properly remove cookies without leftovers?
	if (cookieManager.hasCookie(STATUS_CODE_COOKIE_KEY)) {
		ctx.res.statusCode = Number(
			cookieManager.getCookie(STATUS_CODE_COOKIE_KEY) || DEFAULT_RESPONSE_STATUS_CODE,
		);
		cookieManager.unsetCookie(STATUS_CODE_COOKIE_KEY);
	}

	if (cookieManager.hasCookie(STATUS_MESSAGE_COOKIE_KEY)) {
		ctx.res.statusMessage = cookieManager.getCookie(STATUS_MESSAGE_COOKIE_KEY) || '';
		cookieManager.unsetCookie(STATUS_MESSAGE_COOKIE_KEY);
	}

	const res: NextApiResponse = {};
	if (cookieManager.hasCookie(BODY_COOKIE_KEY)) {
		const resBody = cookieManager.getCookie(BODY_COOKIE_KEY);
		const contentType = cookieManager.getCookie(CONTENT_TYPE_COOKIE_KEY);
		if (contentType === ENCTYPE_APPLICATION_JSON) {
			res.body = deserialize(resBody);
		} else if (contentType === ENCTYPE_APPLICATION_OCTET_STREAM) {
			res.body = deserialize(resBody);
		} else {
			const c = console;
			c.warn('Could not parse response body, returning nothing');
			res.body = null;
		}
		cookieManager.unsetCookie(BODY_COOKIE_KEY);
		cookieManager.unsetCookie(CONTENT_TYPE_COOKIE_KEY);
	}

	const gspResult = (
		typeof options?.fn === 'function'
			? await options.fn(req, res, ctx)
			: {
				props: {},
			}
	);

	if ('props' in gspResult) {
		return {
			...gspResult,
			props: {
				...gspResult.props,
				req,
				res,
			},
		};
	}

	// redirect/not found will be treated as default behavior
	return gspResult;
};
